@@ -4,7 +4,7 @@
 # Ruby Inline is a framework for writing ruby extensions in foreign
 # languages.
 #
-# = SYNOPSIS
+# == SYNOPSIS
 #
 #   require 'inline'
 #   class MyClass
@@ -20,24 +20,39 @@
 #     end
 #   end
 #
-# = DESCRIPTION
+# == DESCRIPTION
 #
 # Inline allows you to write foreign code within your ruby code. It
 # automatically determines if the code in question has changed and
 # builds it only when necessary. The extensions are then automatically
 # loaded into the class/module that defines it.
 #
-# Using the package_inline tool Inline now allows you to package up
-# your inlined object code for distribution to systems without a
-# compiler (read: windows)!
-#
 # You can even write extra builders that will allow you to write
 # inlined code in any language. Use Inline::C as a template and look
 # at Module#inline for the required API.
+#
+# == PACKAGING
+#
+# To package your binaries into a gem, use hoe's INLINE and
+# FORCE_PLATFORM env vars.
+#
+# Example:
+#
+#   rake package INLINE=1
+#
+# or:
+#
+#   rake package INLINE=1 FORCE_PLATFORM=mswin32
+#
+# See hoe for more details.
+#
 
 require "rbconfig"
 require "digest/md5"
 require 'fileutils'
+require 'rubygems'
+
+require 'zentest_mapping'
 
 $TESTING = false unless defined? $TESTING
 
@@ -50,27 +65,62 @@ class CompilationError < RuntimeError; end
 # the current namespace.
 
 module Inline
-  VERSION = '3.6.7'
+  VERSION = '3.8.1'
 
-  WINDOZE  = /win(32|64)/ =~ RUBY_PLATFORM
+  WINDOZE  = /mswin|mingw/ =~ RUBY_PLATFORM
+  RUBINIUS = defined? RUBY_ENGINE
   DEV_NULL = (WINDOZE ? 'nul'      : '/dev/null')
-  RAKE     = (WINDOZE ? 'rake.bat' : 'rake')
   GEM      = (WINDOZE ? 'gem.bat'  : 'gem')
+  RAKE     = if WINDOZE then
+               'rake.bat'
+             elsif RUBINIUS then
+               File.join(Gem.bindir, 'rake')
+             else
+               "#{Gem.ruby} -S rake"
+             end
 
   warn "RubyInline v #{VERSION}" if $DEBUG
 
-  protected
+  def self.register cls
+    registered_inline_classes << cls
+    registered_inline_classes.uniq!
+  end
+
+  def self.registered_inline_classes
+    @@registered_inline_classes ||= []
+  end
+
+  # rootdir can be forced using INLINEDIR variable
+  # if not defined, it should store in user HOME folder
+  #
+  # Under Windows user data can be stored in several locations:
+  #
+  #  HOME
+  #  HOMEDRIVE + HOMEPATH
+  #  APPDATA
+  #  USERPROFILE
+  #
+  # Perform a check in that other to see if the environment is defined
+  # and if so, use it. only try this on Windows.
 
   def self.rootdir
     env = ENV['INLINEDIR'] || ENV['HOME']
 
-    # in case both INLINEDIR and HOME aren't defined, and under Windows
-    # default to HOMEDRIVE + HOMEPATH values
-    env = ENV['HOMEDRIVE'] + ENV['HOMEPATH'] if env.nil? and WINDOZE
+    if env.nil? and WINDOZE then
+      # try HOMEDRIVE + HOMEPATH combination
+      if ENV['HOMEDRIVE'] && ENV['HOMEPATH'] then
+        env = ENV['HOMEDRIVE'] + ENV['HOMEPATH']
+      end
+
+      # no HOMEDRIVE? use APPDATA
+      env = ENV['APPDATA'] if env.nil? and ENV['APPDATA']
+
+      # bummer, still no env? then fall to USERPROFILE
+      env = ENV['USERPROFILE'] if env.nil? and ENV['USERPROFILE']
+    end
 
     if env.nil? then
-      warn "Define INLINEDIR or HOME in your environment and try again"
-      exit 1
+      abort "Define INLINEDIR or HOME in your environment and try again"
     end
 
     unless defined? @@rootdir and env == @@rootdir and test ?d, @@rootdir then
@@ -92,42 +142,47 @@ module Inline
     @@directory
   end
 
+  ##
   # Inline::C is the default builder used and the only one provided by
   # Inline. It can be used as a template to write builders for other
   # languages. It understands type-conversions for the basic types and
-  # can be extended as needed.
+  # can be extended as needed using #add_type_converter, #alias_type_converter
+  # and #remove_type_converter.
 
   class C
 
-    protected unless $TESTING
+    include ZenTestMapping
 
     MAGIC_ARITY_THRESHOLD = 15
     MAGIC_ARITY = -1
 
-    @@type_map = {
-      'char'          => [ 'NUM2CHR',  'CHR2FIX' ],
-      'char *'        => [ 'STR2CSTR', 'rb_str_new2' ],
-      'double'        => [ 'NUM2DBL',  'rb_float_new' ],
-      'int'           => [ 'F'+'IX2INT',  'INT2FIX' ],
-      'long'          => [ 'NUM2INT',  'INT2NUM' ],
-      'unsigned int'  => [ 'NUM2UINT', 'UINT2NUM' ],
-      'unsigned long' => [ 'NUM2UINT', 'UINT2NUM' ],
-      'unsigned'      => [ 'NUM2UINT', 'UINT2NUM' ],
-      'VALUE'         => [ '', '' ],
+    ##
+    # Default C to ruby and ruby to C type map
+
+    TYPE_MAP = {
+      'char'               => [ 'NUM2CHR',        'CHR2FIX'      ],
+
+      'char *'             => [ 'StringValuePtr', 'rb_str_new2'  ],
+
+      'double'             => [ 'NUM2DBL',        'rb_float_new' ],
+
+      'int'                => [ "FI\X2INT",       'INT2FIX'      ],
+      'unsigned int'       => [ 'NUM2UINT',       'UINT2NUM'     ],
+      'unsigned'           => [ 'NUM2UINT',       'UINT2NUM'     ],
+
+      'long'               => [ 'NUM2LONG',       'LONG2NUM'     ],
+      'unsigned long'      => [ 'NUM2ULONG',      'ULONG2NUM'    ],
+
+      'long long'          => [ 'NUM2LL',         'LL2NUM'       ],
+      'unsigned long long' => [ 'NUM2ULL',        'ULL2NUM'      ],
+
+      'off_t'              => [ 'NUM2OFFT',       'OFFT2NUM'     ],
+
+      'VALUE'              => [ '',               ''             ],
       # Can't do these converters because they conflict with the above:
-      # ID2SYM(x), SYM2ID(x), NUM2DBL(x), F\IX2UINT(x)
+      # ID2SYM(x), SYM2ID(x), F\IX2UINT(x)
     }
 
-    def ruby2c(type)
-      raise ArgumentError, "Unknown type #{type.inspect}" unless @@type_map.has_key? type
-      @@type_map[type].first
-    end
-
-    def c2ruby(type)
-      raise ArgumentError, "Unknown type #{type.inspect}" unless @@type_map.has_key? type
-      @@type_map[type].last
-    end
-
     def strip_comments(src)
       # strip c-comments
       src = src.gsub(%r%\s*/\*.*?\*/%m, '')
@@ -148,7 +203,7 @@ module Inline
       sig.gsub!(/\s+/, ' ')
 
       unless defined? @types then
-        @types = 'void|' + @@type_map.keys.map{|x| Regexp.escape(x)}.join('|')
+        @types = 'void|' + @type_map.keys.map{|x| Regexp.escape(x)}.join('|')
       end
 
       if /(#{@types})\s*(\w+)\s*\(([^)]*)\)/ =~ sig then
@@ -189,7 +244,8 @@ module Inline
 
       signature = parse_signature(src, !expand_types)
       function_name = signature['name']
-      method_name = options[:method_name] || function_name
+      method_name = options[:method_name]
+      method_name ||= test_to_normal function_name
       return_type = signature['return']
       arity = signature['arity']
 
@@ -242,6 +298,68 @@ module Inline
       return result if $TESTING
     end # def generate
 
+    ##
+    # Builds a complete C extension suitable for writing to a file and
+    # compiling.
+
+    def generate_ext
+      ext = []
+
+      if @include_ruby_first
+        @inc.unshift "#include \"ruby.h\""
+      else
+        @inc.push "#include \"ruby.h\""
+      end
+
+      ext << @inc
+      ext << nil
+      ext << @src.join("\n\n")
+      ext << nil
+      ext << nil
+      ext << "#ifdef __cplusplus"
+      ext << "extern \"C\" {"
+      ext << "#endif"
+      ext << "  __declspec(dllexport)" if WINDOZE
+      ext << "  void Init_#{module_name}() {"
+      ext << "    VALUE c = rb_cObject;"
+
+      # TODO: use rb_class2path
+      # ext << "    VALUE c = rb_path2class(#{@mod.name.inspect});"
+      ext << @mod.name.split("::").map { |n|
+        "    c = rb_const_get(c, rb_intern(\"#{n}\"));"
+      }.join("\n")
+
+      ext << nil
+
+      @sig.keys.sort.each do |name|
+        method = ''
+        arity, singleton, method_name = @sig[name]
+        if singleton then
+          if method_name == 'allocate' then
+            raise "#{@mod}::allocate must have an arity of zero" if arity > 0
+            ext << "    rb_define_alloc_func(c, (VALUE(*)(VALUE))#{name});"
+            next
+          end
+          method << "    rb_define_singleton_method(c, \"#{method_name}\", "
+        else
+          method << "    rb_define_method(c, \"#{method_name}\", "
+        end
+        method << "(VALUE(*)(ANYARGS))#{name}, #{arity});"
+        ext << method
+      end
+
+      ext << @init_extra.join("\n") unless @init_extra.empty?
+
+      ext << nil
+      ext << "  }"
+      ext << "#ifdef __cplusplus"
+      ext << "}"
+      ext << "#endif"
+      ext << nil
+
+      ext.join "\n"
+    end
+
     def module_name
       unless defined? @module_name then
         module_name = @mod.name.gsub('::','__')
@@ -260,12 +378,14 @@ module Inline
     end
 
     attr_reader :rb_file, :mod
-    if $TESTING then
-      attr_writer :mod
-      attr_accessor :src, :sig, :flags, :libs
-    end
+    attr_writer :mod
+    attr_accessor :src, :sig, :flags, :libs, :init_extra
 
-    public
+    ##
+    # Sets the name of the C struct for generating accessors.  Used with
+    # #accessor, #reader, #writer.
+
+    attr_accessor :struct_name
 
     def initialize(mod)
       raise ArgumentError, "Class/Module arg is required" unless Module === mod
@@ -275,9 +395,9 @@ module Inline
       raise "Couldn't discover caller" if stack.empty?
       real_caller = stack.first
       real_caller = stack[3] if real_caller =~ /\(eval\)/
-      real_caller = real_caller.split(/:/, 3)[0..1]
-      @real_caller = real_caller.join ':'
-      @rb_file = File.expand_path real_caller.first
+      real_caller =~ /(.*):(\d+)/
+      real_caller = $1
+      @rb_file = File.expand_path real_caller
 
       @mod = mod
       @src = []
@@ -287,6 +407,87 @@ module Inline
       @libs = []
       @init_extra = []
       @include_ruby_first = true
+      @inherited_methods = {}
+      @struct_name = nil
+
+      @type_map = TYPE_MAP.dup
+    end
+
+    ##
+    # Adds a #reader and #writer for a C struct member wrapped via
+    # Data_Wrap_Struct.  +method+ is the ruby name to give the accessor,
+    # +type+ is the C type.  Unless the C member name is overridden with
+    # +member+, the method name is used as the struct member.
+    #
+    #   builder.struct_name = 'MyStruct'
+    #   builder.accessor :title,        'char *'
+    #   builder.accessor :stream_index, 'int',   :index
+    #
+    # The latter accesses MyStruct->index via the stream_index method.
+
+    def accessor(method, type, member = method)
+      reader method, type, member
+      writer method, type, member
+    end
+
+    ##
+    # Adds a reader for a C struct member wrapped via Data_Wrap_Struct.
+    # +method+ is the ruby name to give the reader, +type+ is the C type.
+    # Unless the C member name is overridden with +member+, the method
+    # name is used as the struct member.  See #accessor for an example.
+
+    def reader(method, type, member = method)
+      raise "struct name not set for reader #{method} #{type}" unless
+        @struct_name
+
+      c <<-C
+VALUE #{method}() {
+  #{@struct_name} *pointer;
+
+  Data_Get_Struct(self, #{@struct_name}, pointer);
+
+  return #{c2ruby type}(pointer->#{member});
+}
+      C
+    end
+
+    ##
+    # Adds a writer for a C struct member wrapped via Data_Get_Struct.
+    # +method+ is the ruby name to give the writer, +type+ is the C type.
+    # Unless the C member name is overridden with +member+, the method
+    # name is used as the struct member.  See #accessor for an example.
+
+    def writer(method, type, member = method)
+      raise "struct name not set for writer #{method} #{type}" unless
+        @struct_name
+
+      c <<-C
+VALUE #{method}_equals(VALUE value) {
+  #{@struct_name} *pointer;
+
+  Data_Get_Struct(self, #{@struct_name}, pointer);
+
+  pointer->#{member} = #{ruby2c type}(value);
+
+  return value;
+}
+      C
+    end
+
+    ##
+    # Converts ruby type +type+ to a C type
+
+    def ruby2c(type)
+      raise ArgumentError, "Unknown type #{type.inspect}" unless @type_map.has_key? type
+      @type_map[type].first
+    end
+
+    ##
+    # Converts C type +type+ to a ruby type
+
+    def c2ruby(type)
+      raise ArgumentError, "Unknown type #{type.inspect}" unless @type_map.has_key? type
+      @type_map[type].last
     end
 
     ##
@@ -328,47 +529,7 @@ module Inline
         src_name = "#{Inline.directory}/#{module_name}.c"
         old_src_name = "#{src_name}.old"
         should_compare = File.write_with_backup(src_name) do |io|
-          if @include_ruby_first
-            @inc.unshift "#include \"ruby.h\""
-          else
-            @inc.push "#include \"ruby.h\""
-          end
-
-          io.puts
-          io.puts @inc.join("\n")
-          io.puts
-          io.puts @src.join("\n\n")
-          io.puts
-          io.puts
-          io.puts "#ifdef __cplusplus"
-          io.puts "extern \"C\" {"
-          io.puts "#endif"
-          io.puts "  __declspec(dllexport)" if WINDOZE
-          io.puts "  void Init_#{module_name}() {"
-          io.puts "    VALUE c = rb_cObject;"
-
-          # TODO: use rb_class2path
-          io.puts @mod.name.split("::").map { |n|
-            "    c = rb_const_get_at(c,rb_intern(\"#{n}\"));"
-          }.join("\n")
-
-          @sig.keys.sort.each do |name|
-            arity, singleton, method_name = @sig[name]
-            if singleton then
-              io.print "    rb_define_singleton_method(c, \"#{method_name}\", "
-            else
-              io.print "    rb_define_method(c, \"#{method_name}\", "
-            end
-            io.puts  "(VALUE(*)(ANYARGS))#{name}, #{arity});"
-          end
-          io.puts @init_extra.join("\n") unless @init_extra.empty?
-
-          io.puts
-          io.puts "  }"
-          io.puts "#ifdef __cplusplus"
-          io.puts "}"
-          io.puts "#endif"
-          io.puts
+          io.puts generate_ext
         end
 
         # recompile only if the files are different
@@ -387,7 +548,7 @@ module Inline
         if recompile then
 
           hdrdir = %w(srcdir archdir rubyhdrdir).map { |name|
-            dir = Config::CONFIG[name]
+            Config::CONFIG[name]
           }.find { |dir|
             dir and File.exist? File.join(dir, "/ruby.h")
           } or abort "ERROR: Can't find header dir for ruby. Exiting..."
@@ -425,7 +586,7 @@ module Inline
           if $? != 0 then
             bad_src_name = src_name + ".bad"
             File.rename src_name, bad_src_name
-            raise CompilationError, "error executing #{cmd}: #{$?}\nRenamed #{src_name} to #{bad_src_name}"
+            raise CompilationError, "error executing #{cmd.inspect}: #{$?}\nRenamed #{src_name} to #{bad_src_name}"
           end
 
           # NOTE: manifest embedding is only required when using VC8 ruby
@@ -497,8 +658,25 @@ module Inline
     # Registers C type-casts +r2c+ and +c2r+ for +type+.
 
     def add_type_converter(type, r2c, c2r)
-      warn "WAR\NING: overridding #{type} on #{caller[0]}" if @@type_map.has_key? type
-      @@type_map[type] = [r2c, c2r]
+      warn "WAR\NING: overridding #{type} on #{caller[0]}" if @type_map.has_key? type
+      @type_map[type] = [r2c, c2r]
+    end
+
+    ##
+    # Registers C type +alias_type+ as an alias of +existing_type+
+
+    def alias_type_converter(existing_type, alias_type)
+      warn "WAR\NING: overridding #{type} on #{caller[0]}" if
+        @type_map.has_key? alias_type
+
+      @type_map[alias_type] = @type_map[existing_type]
+    end
+
+    ##
+    # Unregisters C type-casts for +type+.
+
+    def remove_type_converter(type)
+      @type_map.delete type
     end
 
     ##
@@ -512,13 +690,19 @@ module Inline
     end
 
     ##
-    # Maps a C constant to ruby (with the same
-    # name). +names_and_types+ is a hash that maps the name of the
-    # constant to its C type.
+    # Maps a C constant to ruby. +names_and_types+ is a hash that maps the
+    # name of the constant to its C type.
+    #
+    #   builder.map_c_const :C_NAME => :int
+    #
+    # If you wish to give the constant a different ruby name:
+    #
+    #   builder.map_c_const :C_NAME => [:int, :RUBY_NAME]
 
     def map_c_const(names_and_types)
       names_and_types.each do |name, typ|
-        self.add_to_init "    rb_define_const(c, #{name.to_s.inspect}, #{c2ruby(typ.to_s)}(#{name}));"
+        typ, ruby_name = Array === typ ? typ : [typ, name]
+        self.add_to_init "    rb_define_const(c, #{ruby_name.to_s.inspect}, #{c2ruby(typ.to_s)}(#{name}));"
       end
     end
 
@@ -590,82 +774,6 @@ module Inline
     end
 
   end # class Inline::C
-  class Packager
-    attr_accessor :name, :version, :summary, :libs_copied, :inline_dir
-
-    def initialize(name, version, summary = '')
-      @name = name
-      @version = version
-      @summary = summary
-      @libs_copied = false
-      @ext = Config::CONFIG['DLEXT']
-
-      # TODO (maybe) put libs in platform dir
-      @inline_dir = File.join "lib", "inline"
-    end
-
-    def package
-      copy_libs
-      generate_rakefile
-      build_gem
-    end
-
-    def copy_libs
-      unless @libs_copied then
-        FileUtils.mkdir_p @inline_dir
-        built_libs = Dir.glob File.join(Inline.directory, "*.#{@ext}")
-        FileUtils.cp built_libs, @inline_dir
-        @libs_copied = true
-      end
-    end
-
-    def generate_rakefile
-      if File.exists? 'Rakefile' then
-        unless $TESTING then
-          STDERR.puts "Hrm, you already have a Rakefile, so I didn't touch it."
-          STDERR.puts "You might have to add the following files to your gemspec's files list:"
-          STDERR.puts "\t#{gem_libs.join "\n\t"}"
-        end
-        return
-      end
-
-      rakefile = eval RAKEFILE_TEMPLATE
-
-      STDERR.puts "==> Generating Rakefile" unless $TESTING
-      File.open 'Rakefile', 'w' do |fp|
-        fp.puts rakefile
-      end
-    end
-
-    def build_gem
-      STDERR.puts "==> Running rake" unless $TESTING or $DEBUG
-
-      cmd = "#{RAKE} package"
-      cmd += "> #{DEV_NULL} 2> #{DEV_NULL}" if $TESTING unless $DEBUG
-
-      if system cmd then
-        unless $TESTING then
-          STDERR.puts
-          STDERR.puts "Ok, you now have a gem in ./pkg, enjoy!"
-        end
-      else
-        STDERR.puts "Calling rake to build the gem failed." unless $TESTING
-      end
-    end
-
-    def gem_libs
-      unless defined? @gem_libs then
-        @gem_libs = Dir.glob File.join(@inline_dir, "*.#{@ext}")
-        files = Dir.glob(File.join('lib', '*')).select { |f| test ?f, f }
-
-        @gem_libs.push(*files)
-        @gem_libs.sort!
-      end
-      @gem_libs
-    end
-
-    RAKEFILE_TEMPLATE = '%[require "rake"\nrequire "rake/gempackagetask"\n\nsummary = #{summary.inspect}\n\nif summary.empty? then\n  STDERR.puts "*************************************"\n  STDERR.puts "*** Summary not filled in, SHAME! ***"\n  STDERR.puts "*************************************"\nend\n\nspec = Gem::Specification.new do |s|\n  s.name = #{name.inspect}\n  s.version = #{version.inspect}\n  s.summary = summary\n\n  s.has_rdoc = false\n  s.files = #{gem_libs.inspect}\n  s.add_dependency "RubyInline", ">= 3.3.0"\n  s.require_path = "lib"\nend\n\ndesc "Builds a gem with #{name} in it"\nRake::GemPackageTask.new spec do |pkg|\n  pkg.need_zip = false\n  pkg.need_tar = false\nend\n]'
-  end # class Packager
 end # module Inline
 
 class Module
@@ -682,6 +790,8 @@ class Module
   # parameter.
 
   def inline(lang = :C, options={})
+    Inline.register self
+
     case options
     when TrueClass, FalseClass then
       warn "WAR\NING: 2nd argument to inline is now a hash, changing to {:testing=>#{options}}" unless options
